namespace PdfMarks {

    public class Main {

        private static bool modify;
        private static string author;
        private static string title;
        private static string subject;
        private static string keywords;
        private static string creator;
        private static string producer;
        private static int mod_date;
        private static string mod_date_s;
        private static int creation_date;
        private static string creation_date_s;

        private const GLib.OptionEntry[] options = {
            { "author", 'a', 0, GLib.OptionArg.STRING,
              ref author, "Author", "AUTHOR" },
            { "title", 't', 0, GLib.OptionArg.STRING,
              ref title, "Title", "TITLE" },
            { "subject", 's', 0, GLib.OptionArg.STRING,
              ref subject, "Subject", "SUBJECT" },
            { "keywords", 'k', 0, GLib.OptionArg.STRING,
              ref keywords, "Keywords", "KEYWORDS" },
            { "creator", 'c', 0, GLib.OptionArg.STRING,
              ref creator, "Creator", "CREATOR" },
            { "producer", 'p', 0, GLib.OptionArg.STRING,
              ref producer, "Producer", "PRODUCER" },
            { "modification-date", 'm', 0, GLib.OptionArg.STRING,
              ref mod_date_s, "Modification date",
              "MODIFICATION DATE" },
            { "creation-date", 'd', 0, GLib.OptionArg.STRING,
              ref creation_date_s, "Creation date", "CREATION DATE" },
            { null }
        };

        private const string CONTEXT =
            "[FILENAMES...] - Edit and show PDF marks";

        private const string DESCRIPTION =
"""With no flags the PDF marks are printed. Dates are always local,
with the format

   YYYY-mm-dd HH:MM:SS

If the PDF has no mark and the correspondig option is missing, the
defaults is used:

   Author:            The user real name.
   Title:             The file name minus the 'pdf' or 'PDF' extension.
   Subject:           'Scanned document'
   Keywords:          'document, scan, copy'
   Creator:           'PDF marks'
   Producer:          The author.
   Modification date: The current local time.
   Creation date:     The current local time.
""";

        private static Poppler.Document get_document(GLib.File file) {
            Poppler.Document document = null;
            try {
                var uri = file.get_uri();
                document = new Poppler.Document.from_file(uri, null);
            } catch (GLib.Error e) {
                error(false, e.message);
            }
            return document;
        }

        private static void print_marks(GLib.File file) {
            var document = get_document(file);
            stdout.printf("%s\n", file.get_basename());
            if (document.author != null)
                stdout.printf("  Author: %s\n", document.author);
            if (document.title != null)
                stdout.printf("  Title: %s\n", document.title);
            if (document.subject != null)
                stdout.printf("  Subject: %s\n", document.subject);
            if (document.keywords != null)
                stdout.printf("  Keywords: %s\n", document.keywords);
            if (document.creator != null)
                stdout.printf("  Creator: %s\n", document.creator);
            if (document.producer != null)
                stdout.printf("  Producer: %s\n", document.producer);
            int64 md = (int64)document.mod_date;
            var mod_date = new GLib.DateTime.from_unix_local(md);
            stdout.printf("  Modification date: %s\n",
                          mod_date.format("%Y-%m-%d %H:%M:%S"));
            int64 cd = (int64)document.creation_date;
            var creation_date = new GLib.DateTime.from_unix_local(cd);
            stdout.printf("  Creation date: %s\n",
                          creation_date.format("%Y-%m-%d %H:%M:%S"));
        }

        private static void set_defaults(Poppler.Document document,
                                         string filename) {
            if (document.author == null)
                document.author = Environment.get_real_name();
            if (document.title == null)
                document.title = filename.replace(".pdf", "");
            if (document.subject == null)
                document.subject = "Scanned document";
            if (document.keywords == null)
                document.keywords = "document, scan, copy";
            if (document.creator == null)
                document.creator = "pdfmarks";
            if (document.producer == null)
                document.producer = author;
        }

        private static void set_marks(GLib.File file) {
            var document = get_document(file);
            set_defaults(document, file.get_basename());
            if (author != null)
                document.author = author;
            if (title != null)
                document.title = title;
            if (subject != null)
                document.subject = subject;
            if (keywords != null)
                document.keywords = keywords;
            if (creator != null)
                document.creator = creator;
            if (producer != null)
                document.producer = producer;
            if (mod_date > 0)
                document.mod_date = mod_date;
            if (creation_date > 0)
                document.creation_date = creation_date;
            try {
                var tmp = GLib.Environment.get_tmp_dir();
                var basename = file.get_basename();
                var path = GLib.Path.build_filename(tmp, basename);
                var saved = GLib.File.new_for_path(path);
                document.save(saved.get_uri());
                saved.move(file,
                           GLib.FileCopyFlags.OVERWRITE|
                           GLib.FileCopyFlags.ALL_METADATA);
            } catch (GLib.Error e) {
                error(false, e.message);
            }
        }

        private static int parse_date(string? date) {
            if (date == null)
                return 0;
            if (date == "")
                return (int)(new DateTime.now_utc()).to_unix();
            var format = "Dates should use the format "+
                "YYYY-mm-dd HH:MM:SS";
            var dt = date.split(" ");
            if (dt.length != 2)
                error(false, format);
            string[] d = dt[0].split("-");
            if (d.length != 3)
                error(false, format);
            string[] t = dt[1].split(":");
            if (t.length != 3)
                error(false, format);
            int year = int.parse(d[0]);
            int month = int.parse(d[1]);
            int day = int.parse(d[2]);
            int hour = int.parse(t[0]);
            int minute = int.parse(t[1]);
            int second = int.parse(t[2]);
            if (year < 1970 ||
                month < 1 || month > 12 ||
                day < 1 || day > 31 ||
                hour < 0 || hour > 23 ||
                minute < 0 || minute > 59 ||
                second < 0 || second > 59)
                error(false, format);
            var datetime = new DateTime.local(year, month, day,
                                              hour, minute, second);
            return (int)datetime.to_unix();
        }

        [PrintfFormat]
        private static void error(bool help, string format, ...) {
            string full_format = "error: " + format + "\n";
            var list = va_list();
            stdout.vprintf(full_format, list);
            if (help)
                stderr.printf("Run ‘pdfmarks --help’ for " +
                              "a list of options.\n");
            Process.exit(1);
        }

        public static void main(string[] args) {
            try {
                var opt = new GLib.OptionContext(CONTEXT);
                opt.set_help_enabled(true);
                opt.add_main_entries(options, null);
                opt.set_description(DESCRIPTION);
                opt.parse(ref args);
            } catch (GLib.Error e) {
                error(true, e.message);
            }

            if (args.length < 2)
                error(true, "Missing file(s)");

            modify = title != null || author != null ||
                keywords != null || mod_date_s != null ||
                creation_date_s != null || creator != null ||
                producer != null;

            mod_date = parse_date(mod_date_s);
            creation_date = parse_date(creation_date_s);

            for (int i = 1; i < args.length; i++) {
                var file = GLib.File.new_for_path(args[i]);
                if (!modify) {
                    print_marks(file);
                } else {
                    set_marks(file);
                }
            }
        }
    }
}
